#! /usr/bin/env python
import argparse
import logging
import signal

from os import kill
from os.path import abspath, dirname, join

from subprocess import PIPE, STDOUT, Popen
from time import sleep
from flask import Flask, redirect, render_template, Response
from jinja2 import FileSystemLoader

from touchandgo.helpers import get_interface, get_lock_diff, LOCKFILE, \
    set_config_dir, get_free_port
from touchandgo.history import History
from touchandgo.lock import Lock
from touchandgo.logger import log_set_up


log = logging.getLogger('touchandgo.proxy')


def serve(py_exec=None):
    log_set_up(True)
    parser = argparse.ArgumentParser()
    parser.add_argument("--python", default="python")
    args = parser.parse_args()

    py_exec = args.python

    app = Flask("touchandgo")
    template_path = join(dirname(__file__), "templates")
    app.jinja_loader = FileSystemLoader(template_path)

    @app.route("/favicon.ico")
    def favicon(self):
        return ""

    @app.route("/<name>")
    @app.route("/<name>/<season>/<episode>")
    def redirect_to(name, season=None, episode=None):
        log.info("Requesting %s %s %s", name, season, episode)
        interface = get_interface()
        path = abspath(__file__)
        dir_path = dirname(path)
        exec_ = join(dir_path, "__init__.py")

        port = "8890"#get_free_port()

        command = '%s %s \"%s\" ' % (py_exec, exec_, name)
        if season is not None:
            command += "%s %s " % (season, episode)
        command += "--daemon --port %s " % port
        log.debug(command)
        process = Popen(command, shell=True, stdout=PIPE, stderr=STDOUT)
        sleep(1)
        while get_lock_diff() < 30:
            sleep(1)
        redirect_url = "http://%s:%s" % (interface, port)
        log.info("redirecting to %s" % redirect_url)
        return redirect(redirect_url)

    @app.route("/l/<name>/<season>/<episode>")
    @app.route("/l/<name>/<season>/<episode>/")
    def list_(name, season, episode):
        url = "http://%s:5000/%s/%s" % (get_interface(), name, season)
        episode = int(episode)

        template = app.jinja_env.get_template("m3u.j2")
        ret = template.render(episodes=range(episode, episode + 10),
                              series=name, season=episode, url=url)
        return Response(response=ret, status=200,
                        mimetype="application/x-mpegurl",
                        content_type="application/x-mpegurl")

    @app.route("/kill")
    def kill_():
        lock = Lock(LOCKFILE)
        kill(lock.get_pid(), signal.SIGQUIT)
        return "OK"

    @app.route("/")
    def main():
        series = []
        keys = []
        items = History.many(sorted="-date,-season,-episode")
        for item in items:
            key = (item.name, item.season is not None)
            if key not in keys:
                series.append(item)
                keys.append(key)

        return render_template("main.html", items=series)


    set_config_dir()
    app.debug = True
    app.run(host="0.0.0.0")


if __name__ == '__main__':
    serve()



#! /usr/bin/env python2
from __future__ import division

import argparse
import thread
import logging

from os import system
from os import _exit
from os.path import join, exists
from time import sleep
from datetime import datetime

from libtorrent import add_magnet_uri, session, storage_mode_t

from touchandgo.constants import STATES
from touchandgo.helpers import is_port_free, get_free_port
from touchandgo.logger import log_set_up
from touchandgo.settings import DEBUG, TMP_DIR, DOWNLOAD_LIMIT, WAIT_FOR_IT, \
    DEFAULT_PORT
from touchandgo.strategy import DefaultStrategy
from touchandgo.stream_server import serve_file
from touchandgo.subtitles import SubtitleDownloader


log = logging.getLogger('touchandgo.download')


class DownloadManager(object):
    strategy_class = DefaultStrategy
    sub_downloader_class = SubtitleDownloader

    def __init__(self, magnet, port=None, sub_lang=None, serve=False):
        self.magnet = magnet
        if port is None:
            port = DEFAULT_PORT
        port = int(port)
        if not is_port_free(port):
            port = get_free_port()

        log.info("[Magnet]: %s [Port]: %s [Sub_lang]: %s [Serve]: %s ",
                 magnet, port, sub_lang, serve)

        self.port = port
        self.serve = serve

        # number of pieces to wait until start streaming
        # we are waiting untill all the first peices are downloaded
        # the biggest file which is supposed to be a video file
        self._video_file = None
        self.callback = serve_file
        self._served_blocks = None
        self.streaming = False

        self.stream_th = None
        self.player_th = None
        self.httpd = None

        self.init_handle()
        self.strategy = self.strategy_class(self)

        if sub_lang is not None:
            self.subtitle = self.sub_downloader_class(sub_lang)
        else:
            self.subtitle = None

    def init_handle(self):
        params = {
            "save_path": TMP_DIR,
            "allocation": storage_mode_t.storage_mode_sparse,
            }
        self.session = session()
        self.handle = add_magnet_uri(self.session, str(self.magnet), params)
        self.handle.set_upload_limit(DOWNLOAD_LIMIT)

    @property
    def status(self):
        return self.handle.status()

    def start(self):
        try:
            self.start_time = datetime.now()
            self.session.listen_on(6881, 6891)
            self.session.start_dht()
            print("Downloading metadata")
            log.info("Downloading metadata")
            while not self.handle.has_metadata():
                print("\n" * 80)
                print self.stats()
                sleep(.5)
            log.info("Starting download")
            self.strategy.initial()

            while True:
                if not self.handle.is_seed():
                    self.strategy.master()
                elif self.strategy.holding_stream:
                    self.strategy.holding_stream = False

                if not self.streaming and self.status.state == 3:
                    self.stream()

                print("\n" * 80)
                print self.stats(DEBUG)
                sleep(1)
        except KeyboardInterrupt:
            if self.httpd is not None:
                self.httpd.shutdown()
            raise KeyboardInterrupt

    @property
    def video_file(self):
        if self._video_file is None:
            self._video_file = self.get_biggest_file()
            log.info("Video File: %s", self._video_file)
        return self._video_file

    def get_biggest_file(self):
        info = self.handle.get_torrent_info()
        biggest_file = ("", 0)
        files = info.files()
        for file_ in files:
            if file_.size > biggest_file[1]:
                biggest_file = [file_.path, file_.size]

        return biggest_file

    def run_vlc(self):
        while not exists(join(TMP_DIR, self.video_file[0])):
            sleep(WAIT_FOR_IT)
        command = "vlc http://localhost:%s -q" % self.port
        if self.subtitle is not None:
            subtitle = self.subtitle.download(self.video_file)
            if subtitle is not None:
                command += " --sub-file %s" % subtitle
        try:
            system(command)
        except KeyboardInterrupt:
            log.debug("Closing VLC")
        _exit(0)

    def stream(self):
        if self.callback is not None and not self.streaming:
            self.streaming = True
            pieces = self.status.pieces
            self._served_blocks = [False for i in range(len(pieces))]
            self.stream_th = thread.start_new_thread(self.callback, (self, ))
            if not self.serve:
                self.player_th = thread.start_new_thread(self.run_vlc, ())

    def block_served(self, block):
        self._served_blocks[block] = True

    def defrag(self):
        numerales = ""
        pieces = self.status.pieces
        for i, piece in enumerate(pieces):
            numeral = "#" if piece else " "
            if self._served_blocks is not None and self._served_blocks[i]:
                numeral = ">"
            numeral += str(self.handle.piece_priority(i))
            numerales += numeral
        return "%s\n" % numerales

    def stats(self, defrag=False):
        status = self.status
        text = ""
        if self._video_file is not None:
            text += "Serving %s on http://localhost:%s\n" % (self.video_file[0],
                                                             self.port)
        if defrag:
            text += self.defrag()
        text += '%s %.2f%% complete ' % (STATES[status.state],
                                         status.progress * 100)
        text += '(down: %.1f kB/s up: %.1f kB/s peers: %d)\n' % \
            (status.download_rate / 1000, status.upload_rate / 1000,
             status.num_peers)
        text += "Elapsed Time %s" % (datetime.now() - self.start_time)
        return text


def main():
    log_set_up()
    parser = argparse.ArgumentParser()

    parser.add_argument("magnet", nargs='?', default=None)
    parser.add_argument("--sub", nargs='?', default=None)
    parser.add_argument("--serve", action="store_true")
    parser.add_argument("--port", "-p", default="8888")

    args = parser.parse_args()

    manager = DownloadManager(args.magnet, port=args.port, serve=args.serve,
                              sub_lang=args.sub)
    manager.start()

if __name__ == "__main__":
    main()



import logging


log = logging.getLogger('touchandgo.strategy')


class DefaultStrategy(object):
    def __init__(self, manager):
        self.manager = manager
        self.piece_st = 4
        self.holding_stream = True
        self.handle = manager.handle
        self.chunks_strat = 10
        self.last_pieces_count = 5

    def initial(self):
        self.handle.set_sequential_download(True)
        status = self.handle.status()

        #for i in range(self.piece_st):
        #    self.handle.piece_priority(i, 7)
        #    self.handle.set_piece_deadline(i, 3000)

        last_piece = len(status.pieces)
        for i in range(last_piece-self.last_pieces_count, last_piece):
            self.handle.piece_priority(i, 7)
            self.handle.set_piece_deadline(i, 1000)

        self.chunks_strat = len(status.pieces) / 30

    def block_requested(self, block_requested):
        #log.debug("block requested %s" % block_requested)
        if not self.holding_stream:
            self.move_strategy(block_requested)

    def master(self):
        status = self.handle.status()
        pieces = status.pieces
        first_n = pieces[:self.piece_st] + pieces[-5:]

        if all(first_n):
            self.handle.set_sequential_download(False)
            pieces_strat = pieces[self.piece_st:self.piece_st + self.chunks_strat]
            if self.holding_stream or all(pieces_strat):
                if not self.holding_stream:
                    self.piece_st += self.chunks_strat
                else:
                    self.holding_stream = False
                    self.manager.stream()
                self.move_strategy(self.piece_st)

    def reset_priorities(self):
        for i in range(len(self.handle.status().pieces)):
            self.handle.piece_priority(i, 1)

    def move_strategy(self, first_block):
        log.debug("moving strategy %s" % first_block)
        self.reset_priorities()
        self.piece_st = first_block

        self.handle.set_sequential_download(True)
        status = self.handle.status()
        last_piece = len(status.pieces) - 1

        last_chunks = first_block + self.chunks_strat
        if last_chunks > last_piece:
            last_chunks = last_piece

        for i in range(first_block, last_chunks):
            self.handle.piece_priority(i, 7)
            self.handle.set_piece_deadline(i, 10000)

        last_chunks = first_block + self.chunks_strat * 2
        if last_chunks > last_piece:
            last_chunks = last_piece
        for i in range(first_block+self.chunks_strat, last_chunks):
            self.handle.piece_priority(i, 3)
            #self.handle.set_piece_deadline(i, 20000)
